#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Manage DNS/DHCP entries by utilizing the web services at komcenter.ethz.ch.
#
# (c) 2009 - 2012, Steven Armstrong, ETH Zurich
#

__version__ = '0.3.1'

import sys
import os
import re
import xml.etree.ElementTree as ET


import logging
logging.basicConfig(level=logging.ERROR, format='%(levelname)s: %(message)s')


from ethz import cli
from ethz.cli import util
from ethz.cli import input
from ethz.cli import output
from ethz.util import httpclient
from ethz.util import get_username_password
from ethz.util import format_mac
from ethz.api import komcenter
from ethz import network
from ethz import Error

import configobj



class IpCtl(cli.Command):
    """Usage:
        ${name} COMMAND [ARGS...]
        ${name} help COMMAND

    Manage DNS/DHCP entries by utilizing the web services at komcenter.ethz.ch.

    ${option_list}
    ${command_list}
    ${help_list}
    """
    name = 'ipctl'
    option_interactive = True
    option_debug = True

    default_http_headers = {'Content-Type': 'text/xml'}

    _http = None
    @property
    def http(self):
        if not self._http:
            self._http = httpclient.get_http_client(username=self.options.user, host=komcenter.hostname)
        return self._http

    _komcenter = None
    @property
    def komcenter(self):
        if not self._komcenter:
            username,password = get_username_password(username=self.options.user, host=komcenter.hostname)
            self._komcenter = komcenter.Client(username=username, password=password)
        return self._komcenter

    _config = None
    @property
    def config(self):
        if not self._config:
            config_file = os.path.expanduser(self.options.config)
            if os.path.isfile(config_file):
                self._config = configobj.ConfigObj(config_file)
            else:
                self._config = {}
        return self._config

    def get_optparser(self):
        parser = super(IpCtl, self).get_optparser()
        #parser.add_option('-c', '--config', type='string', dest='config', default='~/.ipctl.conf',
        #    help='Path to config file, default: ~/.hostctl.conf')
        parser.add_option('-u', '--user', type='string', dest='user', default=None,
            help='the username for authenticating against webservices')
        return parser


    #---------- Commands ----------

    _autodetect_patterns = {
        # 127.0.0.1/host.ethz.ch
        'ip/hostname': re.compile(r'^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/(?=.{1,255}$)[0-9a-z+](?:(?:[0-9a-z]|\b-){0,61}[0-9a-z])?(?:\.[0-9a-z](?:(?:[0-9a-z]|\b-){0,61}[0-9a-z])?)*\.?$', re.IGNORECASE),
        # 127.0.0.1
        'ip': re.compile(r'^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$'),
        'ipv6': re.compile(r'^\s*((([0-9A-Fa-f]{1,4}:){7}([0-9A-Fa-f]{1,4}|:))|(([0-9A-Fa-f]{1,4}:){6}(:[0-9A-Fa-f]{1,4}|((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3})|:))|(([0-9A-Fa-f]{1,4}:){5}(((:[0-9A-Fa-f]{1,4}){1,2})|:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3})|:))|(([0-9A-Fa-f]{1,4}:){4}(((:[0-9A-Fa-f]{1,4}){1,3})|((:[0-9A-Fa-f]{1,4})?:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(([0-9A-Fa-f]{1,4}:){3}(((:[0-9A-Fa-f]{1,4}){1,4})|((:[0-9A-Fa-f]{1,4}){0,2}:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(([0-9A-Fa-f]{1,4}:){2}(((:[0-9A-Fa-f]{1,4}){1,5})|((:[0-9A-Fa-f]{1,4}){0,3}:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(([0-9A-Fa-f]{1,4}:){1}(((:[0-9A-Fa-f]{1,4}){1,6})|((:[0-9A-Fa-f]{1,4}){0,4}:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(:(((:[0-9A-Fa-f]{1,4}){1,7})|((:[0-9A-Fa-f]{1,4}){0,5}:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:)))(%.+)?\s*$'),
        # host.ethz.ch | ethz.ch | +.cbrg.ethz.ch (allow leading + as that's what ID-KOM abuses to address a subdomain)
        'hostname': re.compile(r'^(?=.{1,255}$)[0-9a-z+](?:(?:[0-9a-z]|\b-){0,61}[0-9a-z])?(?:\.[0-9a-z](?:(?:[0-9a-z]|\b-){0,61}[0-9a-z])?)*\.?$', re.IGNORECASE),
        # 00:1B:FC:A2:C3:9F | 00-1B-FC-A2-C3-9F
        'mac': re.compile(r'^(?:[a-f0-9]{2}[\:-]){5}[a-f0-9]{2}$', re.IGNORECASE),
    }

    _ip_add_properties = ('ip', 'ipv6', 'fqName', 'forward', 'reverse', 'ttl', 'dhcp', 'dhcpMac', 'ddns', 'isgGroup')
    def do_ip_add(self, subcmd, opts, *args):
        """add new DNS entry for FQDN in SUBNET

        Prints the allocated IP to stdout.

        Usage:
            ${name} ${cmd_name} subnet fqName [attribute-1:value], ..., attribute-n:value

        Where available attributes are:
            subnet*          Subnet in which to create the new DNS entry.
            fqName*          Full qualified host name
            ip               IPv4 address. If ommited the next free IP in the subnet is used.
            ipv6             IPv6 address.
            forward          Create a DNS A record
            reverse          Create a DNS PTR record
            ttl              Time to live
            dhcp             Use ID-KOM dhcp server?
            dhcpMac          MAC address for ID-KOM dhcp server
            ddns             Use ID-KOM dynamic DNS
            isgGroup         The ISG group this record belongs to

        A * denotes a required field.

        Attribute names can be abbreviated as long as a unique prefix is used.
        You will be prompted to select values for missing required attributes.
        """
        assert args, 'Missing required arguments: subnet fqName'
        assert len(args) >= 2, 'Wrong number of arguments. Expected at least 2, got %d' % len(args)

        args = list(args)
        subnet = args[0]
        fqName = args[1]

        properties = util.parse_properties(args[2:], self._ip_add_properties)
        properties['fqName'] = fqName

        automatic_ip_allocation = False
        if not 'ip' and not 'ipv6' in properties:
            freeips = self._get_freeip(subnet, xml=False)
            if freeips:
                properties['ip'] = freeips[0]
                automatic_ip_allocation = True
            else:
                raise Error('No free IPs in subnet %s' % subnet)

        assert 'ip' in properties or 'ipv6' in properties, 'Missing required property \'ip\' or \'ipv6\''
        assert not ('ip' in properties and 'ipv6' in properties), '\'ip\' and \'ipv6\' can\'t both be set'
        assert 'fqName' in properties, 'Missing required property \'fqName\''

        if 'dhcpMac' in properties:
            if not 'dhcp' in properties or not properties['dhcp']:
                logging.info('dhcpMac given, but dhcp not. Implicitly setting \'dhcp\' to: Y')
                properties['dhcp'] = 'Y'

        if not 'isgGroup' in properties or not properties['isgGroup']:
            # prompt for isg
            isgs = self._nac_isgs()
            labels = ['Id', 'Name']
            entries = [[x['id'], x['name'], x['email']] for x in isgs]
            isg = input.choose_one(isgs, fields=('id','name'), labels=labels, intro='Please select a isg:')
            properties['isgGroup'] = isg['name']
            logging.info('Setting \'isgGroup\' to: %s', isg['name'])

        logging.debug('properties: %s', properties)

        uri = komcenter.uri('/netcenter/rest/nameToIP')
        headers = self.default_http_headers.copy()
        body = ['<insert><nameToIP>']
        for k,v in properties.items():
            if k in self._ip_add_properties:
                body.append('<%s>%s</%s>' % (k,v,k))
        body.append('</nameToIP></insert>')
        body = ''.join(body)
        logging.debug('uri: %s', uri)
        logging.debug('body: %s', body)

        response, content = self.http.request(uri, method='POST', headers=headers, body=body, auth_without_challenge=True)
        logging.debug('response: %s', response)
        logging.debug('content: %s', content)

        tree = komcenter.get_tree_or_raise_error(response, content)
        # If we get here, creation was successfull

        if 'ip' in properties:
          logging.info('Successfully created DNS/DHCP entry for %(fqName)s %(ip)s' % properties)
        else:
          logging.info('Successfully created DNS/DHCP entry for %(fqName)s %(ipv6)s' % properties)

        if automatic_ip_allocation and not self.options.quiet:
            # Tell the user which ip was allocated for him
            print properties['ip']


    def do_ip_delete(self, subcmd, opts, *args):
        """delete the DNS entry of the given IP address

        Usage:
            ${name} ${cmd_name} key value

        Where key is one of:
            ip               delete entry by ip
            ip/hostname      delete entry for given ip/hostname pair
            hostname         resolve ip from hostname,
                             then delete entry by ip/hostname pair
            mac              resolve ip and hostname from mac,
                             then delete entry by ip/hostname pair

        Key names can be abbreviated as long as a unique prefix is used.

        If only a value is given, the key is guessed by inspecting the value:
            ${name} ${cmd_name} 129.132.12.88 -> ${name} ${cmd_name} ip 129.132.12.88
            ${name} ${cmd_name} 129.132.12.88/foo.inf.ethz.ch -> ${name} ${cmd_name} ip/hostname 129.132.12.88/foo.inf.ethz.ch
            ${name} ${cmd_name} foo.inf.ethz.ch -> ${name} ${cmd_name} hostname foo.inf.ethz.ch
            ${name} ${cmd_name} 00:DA:DE:BA:BE:00 -> ${name} ${cmd_name} mac 00:DA:DE:BA:BE:00

        ${cmd_option_list}
        """
        assert args, 'Missing required arguments: key, value'
        assert len(args) <= 2, 'Wrong number of arguments. Expected 1 or 2, got %d' % len(args)

        (attribute, value) = self._attribute_value_from_arguments(args)
        host_info = self._host_info_from_attribute_value(attribute, value)
        if attribute == 'ip':
            logging.debug('Deleting all entries associated with ip: {ip}'.format(**host_info))
            uri = komcenter.uri('/netcenter/rest/nameToIP/{ip}'.format(**host_info))
        else:
            logging.debug('Deleting entry for ip/hostname: {ip}/{hostname}'.format(**host_info))
            uri = komcenter.uri('/netcenter/rest/nameToIP/{ip}/{hostname}'.format(**host_info))
        logging.debug('uri: %s', uri)
        headers = self.default_http_headers.copy()
        response, content = self.http.request(uri, method='DELETE', headers=headers, auth_without_challenge=True)
        logging.debug('response: %s', response)
        logging.debug('content: %s', content)

        tree = komcenter.get_tree_or_raise_error(response, content)
        # If we get here, deletion was successfull
        logging.info('Successfully deleted entry for %s', value)


    @cli.option('-x', '--xml', action='store_true',
                  help='output raw xml')
    def do_ip_get(self, subcmd, opts, *args):
        """get the DNS/DHCP entry for the given IP address

        Usage:
            ${name} ${cmd_name} key value

        Where key is one of:
            ip               get entry by ip
            ip/hostname      get entry for given ip/hostname pair
            hostname         resolve ip from hostname,
                             then get entry by ip/hostname pair
            mac              resolve ip and hostname from mac,
                             then get entry by ip/hostname pair

        Key names can be abbreviated as long as a unique prefix is used.

        If only a value is given, the key is guessed by inspecting the value:
            ${name} ${cmd_name} 129.132.12.88 -> ${name} ${cmd_name} ip 129.132.12.88
            ${name} ${cmd_name} 129.132.12.88/foo.inf.ethz.ch -> ${name} ${cmd_name} ip/hostname 129.132.12.88/foo.inf.ethz.ch
            ${name} ${cmd_name} foo.inf.ethz.ch -> ${name} ${cmd_name} hostname foo.inf.ethz.ch
            ${name} ${cmd_name} 00:DA:DE:BA:BE:00 -> ${name} ${cmd_name} mac 00:DA:DE:BA:BE:00

        ${cmd_option_list}
        """
        assert args, 'Missing required arguments: key, value'
        assert len(args) <= 2, 'Wrong number of arguments. Expected 1 or 2, got %d' % len(args)

        (attribute, value) = self._attribute_value_from_arguments(args)
        host_info = self._host_info_from_attribute_value(attribute, value)
        if attribute == 'ip':
            logging.debug('Getting all entries associated with ip: {ip}'.format(**host_info))
            uri = komcenter.uri('/netcenter/rest/nameToIP/{ip}'.format(**host_info))
        else:
            logging.debug('Getting entry for ip/hostname: {ip}/{hostname}'.format(**host_info))
            uri = komcenter.uri('/netcenter/rest/nameToIP/{ip}/{hostname}'.format(**host_info))
        logging.debug('uri: %s', uri)
        headers = self.default_http_headers.copy()
        response, content = self.http.request(uri, method='GET', headers=headers, auth_without_challenge=True)
        logging.debug('response: %s', response)
        logging.debug('content: %s', content)

        if opts.xml:
            return content
        else:
            tree = komcenter.get_tree_or_raise_error(response, content)
            usedIps = tree.findall('usedIp')
            if usedIps:
                record_count = len(usedIps)
                for index,usedIp in enumerate(usedIps):
                    record = komcenter.record_to_dict(usedIp)
                    for key,value in record.items():
                        print '%s: %s' % (key, value)
                    if index < record_count-1:
                       print '--'
            else:
                # no entry found, exit non zero for easier use in scripts
                sys.exit(1)


    _ip_set_attributes = ('ip/hostname', 'hostname', 'mac')
    _ip_set_properties = ('ip', 'fqName', 'forward', 'reverse', 'ttl', 'dhcp', 'dhcpMac', 'ddns', 'isgGroup')
    def do_ip_set(self, subcmd, opts, *args):
        """set/change properties of an existing DNS/DHCP IP entry

        Usage:
            ${name} ${cmd_name} [key] value attribute1:value attribute2:value ...

        Where available attributes are:
            ip               IPv4 address
            fqName           Full qualified host name
            forward          Create a DNS A record
            reverse          Create a DNS PTR record
            ttl              Time to live
            dhcp             Use ID-KOM dhcp server?
            dhcpMac          MAC address for ID-KOM dhcp server
            ddns             Use ID-KOM dynamic DNS
            isgGroup         The ISG group this record belongs to

        A * denotes a required field.

        Attribute names can be abbreviated as long as a unique prefix is used.
        You will be prompted to select values for missing required attributes.

        Key must be one of:
            ip               change entry by ip
            ip/hostname      change entry for given ip/hostname pair
            hostname         resolve ip from hostname,
                             then change entry by ip/hostname pair
            mac              resolve ip and hostname from mac,
                             then change entry by ip/hostname pair

        Key names can be abbreviated as long as a unique prefix is used.

        If only a value is given, the key is guessed by inspecting the value:
            ${name} ${cmd_name} 129.132.12.88 -> ${name} ${cmd_name} ip 129.132.12.88
            ${name} ${cmd_name} 129.132.12.88/foo.inf.ethz.ch -> ${name} ${cmd_name} ip/hostname 129.132.12.88/foo.inf.ethz.ch
            ${name} ${cmd_name} foo.inf.ethz.ch -> ${name} ${cmd_name} hostname foo.inf.ethz.ch
            ${name} ${cmd_name} 00:DA:DE:BA:BE:00 -> ${name} ${cmd_name} mac 00:DA:DE:BA:BE:00
        """
        assert args, 'Missing required arguments'
        args = list(args)

        properties = util.parse_properties(args, self._ip_set_properties)
        if 'dhcpMac' in properties:
            if not 'dhcp' in properties or not properties['dhcp']:
                logging.info('dhcpMac given, but dhcp not. Implicitly setting \'dhcp\' to: Y')
                properties['dhcp'] = 'Y'
        logging.debug('properties: %s', properties)

        # construct the uri to the resource
        (attribute, value) = self._attribute_value_from_arguments(args)
        host_info = self._host_info_from_attribute_value(attribute, value)
        if attribute == 'ip':
            get_uri = komcenter.uri('/netcenter/rest/nameToIP/{ip}'.format(**host_info))
        else:
            get_uri = komcenter.uri('/netcenter/rest/nameToIP/{ip}/{hostname}'.format(**host_info))
        logging.debug('get_uri: %s', get_uri)

        # GET current entry
        headers = self.default_http_headers.copy()
        response, content = self.http.request(get_uri, method='GET', headers=headers, auth_without_challenge=True)
        logging.debug('response: %s', response)
        logging.debug('content: %s', content)
        tree = komcenter.get_tree_or_raise_error(response, content)
        usedIps = tree.findall('usedIp')
        logging.debug('len(usedIps): %s', len(usedIps))
        record = None
        if usedIps is None or len(usedIps) == 0:
            raise Error('Can not update nonexistent entry for: %s' % args)
        elif len(usedIps) == 1:
            record = komcenter.record_to_dict(usedIps[0])
        elif len(usedIps) > 1:
            # choose record from list
            records = [komcenter.record_to_dict(usedIp) for usedIp in usedIps]
            labels = ['IP', 'Hostname']
            record = input.choose_one(records, fields=('ip', 'fqname'), labels=labels, intro='There is more then one hostname assigned to this ip, please select the one to change:')

        host_info['hostname'] = record['fqname']
        logging.debug('record: %s', record)

        # update it from given properties
        record.update(properties)

        # PUT it back
        body = ['<modify><nameToIP>']
        for k,v in record.items():
            if k in self._ip_set_properties:
                if v:
                    body.append('<%s>%s</%s>' % (k,v,k))
                else:
                    body.append('<%s />' % k)
        body.append('</nameToIP></modify>')
        body = ''.join(body)
        put_uri = komcenter.uri('/netcenter/rest/nameToIP/{ip}/{hostname}'.format(**host_info))
        logging.debug('put_uri: %s', put_uri)
        logging.debug('body: %s', body)
        response, content = self.http.request(put_uri, method='PUT', headers=headers, body=body, auth_without_challenge=True)
        logging.debug('response: %s', response)
        logging.debug('content: %s', content)

        tree = komcenter.get_tree_or_raise_error(response, content)
        # If we get here, creation was successfull
        logging.info('Successfully changed DNS/DHCP entry for {ip}/{hostname}'.format(**host_info))


    @cli.option('-x', '--xml', action='store_true',
                  help='output raw xml')
    @cli.option('-1', '--one', action='store_true',
                  help='print just the first free IP instead of all available')
    def do_ip_free(self, subcmd, opts, *args):
        """search for free IP addresses in given subnet

        Usage:
            ${name} ${cmd_name} <subnet>

        ${cmd_option_list}
        """
        assert args, 'Missing required argument: subnet'
        assert len(args) == 1, 'Wrong number of arguments. Expected 1, got %d' % len(args)

        subnet = args[0]

        result = self._get_freeip(subnet, xml=opts.xml)

        if opts.xml:
            print result
        else:
            if opts.one:
                print result[0]
            else:
                for ip in result:
                    print ip



    @cli.option('-x', '--xml', action='store_true',
                  help='output raw xml')
    def do_ip_used(self, subcmd, opts, *args):
        """search for used IP addresses in given subnet

        Usage:
            ${name} ${cmd_name} <subnet>

        ${cmd_option_list}
        """
        assert args, 'Missing required argument: subnet'
        assert len(args) == 1, 'Wrong number of arguments. Expected 1, got %d' % len(args)

        subnet = args[0]

        uri = komcenter.uri('/netcenter/rest/nameToIP/usedIps/v4/%s' % subnet)
        headers = self.default_http_headers.copy()
        logging.debug('uri: %s', uri)

        response, content = self.http.request(uri, method='GET', headers=headers, auth_without_challenge=True)
        logging.debug('response: %s', response)
        logging.debug('content: %s', content)

        tree = komcenter.get_tree_or_raise_error(response, content)

        if opts.xml:
            print content
        else:
            freeips = tree.getchildren()
            for freeip in freeips:
                print freeip.find('ip').text



    _alias_add_properties = ('ttl',)
    def do_alias_add(self, subcmd, opts, *args):
        """add new ALIAS for HOST

        Usage:
            ${name} ${cmd_name} host alias [attribute-1:value], ..., attribute-n:value

        Where available attributes are:
            ttl              Time to live

        A * denotes a required field.

        Attribute names can be abbreviated as long as a unique prefix is used.
        You will be prompted to select values for missing required attributes.
        """
        assert args, 'Missing required arguments: host alias'
        assert len(args) >= 2, 'Wrong number of arguments. Expected at least 2, got %d' % len(args)

        args = list(args)
        host = args[0]
        alias = args[1]

        properties = util.parse_properties(args[2:], self._alias_add_properties)
        properties['fqName'] = alias
        properties['hostName'] = host

        #properties = self._alias_create(properties)
        logging.debug('properties: %s', properties)

        uri = komcenter.uri('/netcenter/rest/alias')
        headers = self.default_http_headers.copy()
        body = ['<insert><alias>']
        for k,v in properties.items():
            if v:
                body.append('<%s>%s</%s>' % (k,v,k))
            else:
                body.append('<%s />' % k)
        body.append('</alias></insert>')
        body = ''.join(body)
        logging.debug('uri: %s', uri)
        logging.debug('body: %s', body)

        response, content = self.http.request(uri, method='POST', headers=headers, body=body, auth_without_challenge=True)
        logging.debug('response: %s', response)
        logging.debug('content: %s', content)

        tree = komcenter.get_tree_or_raise_error(response, content)
        # If we get here, creation was successfull
        logging.info('Successfully created DNS entry for %(fqName)s -> %(hostName)s' % properties)



    def do_alias_delete(self, subcmd, opts, *args):
        """delete the DNS entry of the given Alias

        Usage:
            ${name} ${cmd_name} alias
        """
        assert args, 'Missing required argument: alias'
        assert len(args) == 1, 'Wrong number of arguments. Expected 1, got %d' % len(args)

        alias = args[0]

        uri = komcenter.uri('/netcenter/rest/alias/{0}'.format(alias))
        headers = self.default_http_headers.copy()
        logging.debug('uri: %s', uri)

        response, content = self.http.request(uri, method='DELETE', headers=headers, auth_without_challenge=True)
        logging.debug('response: %s', response)
        logging.debug('content: %s', content)

        tree = komcenter.get_tree_or_raise_error(response, content)
        # If we get here, deletion was successfull
        logging.info('Successfully deleted Alias %s', alias)



    @cli.option('-x', '--xml', action='store_true',
                  help='output raw xml')
    def do_alias_get(self, subcmd, opts, *args):
        """get the DNS entry for the given Alias

        Usage:
            ${name} ${cmd_name} <alias>

        Where argument are:
            alias            Full qualified alias name

        ${cmd_option_list}
        """
        assert args, 'Missing required argument: alias'
        assert len(args) == 1, 'Wrong number of arguments. Expected 1, got %d' % len(args)

        alias = args[0]

        uri = komcenter.uri('/netcenter/rest/alias/{0}'.format(alias))
        headers = self.default_http_headers.copy()
        logging.debug('uri: %s', uri)

        response, content = self.http.request(uri, method='GET', headers=headers, auth_without_challenge=True)
        logging.debug('response: %s', response)
        logging.debug('content: %s', content)

        if opts.xml:
            # FIXME: return output instead of printing everywhere
            #return content
            print(content)
        else:
            tree = komcenter.get_tree_or_raise_error(response, content)
            alias = tree.find('alias')
            if alias is not None:
                record = komcenter.record_to_dict(alias)
                for key,value in record.items():
                    print('{0}: {1}'.format(key, value))



    _alias_set_properties = ('hostName', 'ttl')
    def do_alias_set(self, subcmd, opts, *args):
        """set/change properties of an existing DNS alias entry

        Usage:
            ${name} ${cmd_name} alias [attribute-1:value], ..., attribute-n:value

        Where available attributes are:
            hostName         Full qualified host name
            ttl              Time to live

        A * denotes a required field.

        Attribute names can be abbreviated as long as a unique prefix is used.
        You will be prompted to select values for missing required attributes.
        """
        assert args, 'Missing required arguments: alias'
        args = list(args)
        alias = args[0]

        properties = util.parse_properties(args[1:], self._alias_set_properties)
        properties['fqName'] = alias
        logging.debug('properties: %s', properties)

        uri = komcenter.uri('/netcenter/rest/alias/{0}'.format(alias))
        headers = self.default_http_headers.copy()
        body = ['<modify><alias>']
        for k,v in properties.items():
            if v:
                body.append('<%s>%s</%s>' % (k,v,k))
            else:
                body.append('<%s />' % k)
        body.append('</alias></modify>')
        body = ''.join(body)
        logging.debug('uri: %s', uri)
        logging.debug('body: %s', body)

        response, content = self.http.request(uri, method='PUT', headers=headers, body=body, auth_without_challenge=True)
        logging.debug('response: %s', response)
        logging.debug('content: %s', content)

        tree = komcenter.get_tree_or_raise_error(response, content)
        # If we get here, the operation was successfull
        logging.info('Successfully changed DNS/DHCP entry for %s', alias)



    # FIXME: This should really be in the ip context. e.g. ipctl ip aliases host.domain.tld
    @cli.option('-x', '--xml', action='store_true',
                  help='output raw xml')
    def do_alias_all(self, subcmd, opts, *args):
        """search for all Aliases for Host

        Usage:
            ${name} ${cmd_name} ip|hostname

        ${cmd_option_list}
        """
        assert args, 'Missing required argument: hostname'
        assert len(args) == 1, 'Wrong number of arguments. Expected 1, got %d' % len(args)

        ip_or_hostname = args[0]

        if self._autodetect_patterns['hostname'].search(ip_or_hostname):
            hostname = ip_or_hostname
        elif self._autodetect_patterns['ip'].search(ip_or_hostname) or \
                self._autodetect_patterns['ipv6'].search(ip_or_hostname):
            logging.info('resolving hostname for ip address %s', ip_or_hostname)
            hostname = network.get_hostname(ip_or_hostname, komcenter_client=self.komcenter)
        else:
            raise Error('Expected ip address or hostname, got %s' % ip_or_hostname)


        uri = komcenter.uri('/netcenter/rest/alias/hostName/%s' % hostname)
        headers = self.default_http_headers.copy()
        logging.debug('uri: %s', uri)

        response, content = self.http.request(uri, method='GET', headers=headers, auth_without_challenge=True)
        logging.debug('response: %s', response)
        logging.debug('content: %s', content)

        tree = komcenter.get_tree_or_raise_error(response, content)

        if opts.xml:
            print content
        else:
            aliases = tree.findall('alias')
            if aliases:
                record_count = len(aliases)
                for index,alias in enumerate(aliases):
                    record = komcenter.record_to_dict(alias)
                    for key,value in record.items():
                        print '%s: %s' % (key, value)
                    if index < record_count-1:
                       print '--'



    @cli.alias('q')
    @cli.alias('exit')
    def _do_quit(self, argv):
        """hidden quit command to leave interactive shell"""
        sys.exit(0)



    #---------- Utilities ----------


    _ip_attributes = ('ip', 'ip/hostname', 'hostname', 'mac')
    def _attribute_value_from_arguments(self, args, attributes=None):
        attributes = attributes or self._ip_attributes
        attribute = None
        if len(args) == 1:
            value = args[0]
            for name, pattern in self._autodetect_patterns.items():
                if pattern.search(value):
                    if name == 'ipv6':
                        name = 'ip'
                    attribute = name
                    break
            assert attribute, 'Failed to detect attribute name for value: {0}'.format(value)
        elif len(args) == 2:
            attribute = util.get_canonical_name(args[0], attributes, label='attribute')
            value = args[1]

        logging.debug('attribute: %s', attribute)
        logging.debug('value: %s', value)
        return attribute, value


    def _host_info_from_attribute_value(self, attribute, value, attributes=None):
        attributes = attributes or self._ip_attributes
        host_info = dict.fromkeys(attributes)
        host_info[attribute] = value
        if attribute == 'ip/hostname':
            host_info['ip'],host_info['hostname'] = value.split('/')
        if attribute == 'hostname':
            logging.info('resolving ip address for %s %s', attribute, value)
            host_info['ip'] = network.get_ip(value, komcenter_client=self.komcenter)
        if attribute == 'mac':
            # komcenter suddenly on likes mac addresses delimited with -
            value = format_mac(value, '-')
            logging.info('getting host info for %s %s', attribute, value)
            _host_info = network.get_host_info(value, komcenter_client=self.komcenter)
            if _host_info:
                host_info['ip'] = _host_info['ip']
                host_info['hostname'] = _host_info['ip_name']
            else:
                logging.warning('failed to get host info for %s %s', attribute, value)
                logging.info('resolving ip address for %s %s', attribute, value)
                host_info['ip'] = network.get_ip(value, komcenter_client=self.komcenter)
                logging.info('resolving hostname for %s %s', attribute, value)
                host_info['hostname'] = network.get_hostname(value, komcenter_client=self.komcenter)
        logging.debug('host_info: %s', host_info)
        return host_info


    def _get_freeip(self, subnet, xml=False):
        uri = komcenter.uri('/netcenter/rest/nameToIP/freeIps/v4/%s' % subnet)
        headers = self.default_http_headers.copy()
        logging.debug('uri: %s', uri)

        response, content = self.http.request(uri, method='GET', headers=headers, auth_without_challenge=True)
        logging.debug('response: %s', response)
        logging.debug('content: %s', content)

        tree = komcenter.get_tree_or_raise_error(response, content)

        if xml:
            return content
        else:
            freeips = [freeip.find('ip').text for freeip in tree.getchildren()]
            return freeips



    # FIXME: This is a copy of nacctl.py:_nac_isgs. Refactor it out into a seperate module
    def _nac_isgs(self, list_all=False):
        """get a list of isgs from the webservice"""
        """
<isg>
    <id>4136303</id>
    <name>inf-icos</name>
    <email>support@inf.ethz.ch</email>
    <vpzs>
        <vpz><name>inf-icos</name></vpz>
    </vpzs>
</isg>
        """
        if list_all:
            uri = komcenter.uri('/netcenter/rest/user/listAllIsgs')
        else:
            uri = komcenter.uri('/netcenter/rest/user/listIsgByUser')
        headers = self.default_http_headers.copy()
        logging.debug('uri: %s', uri)

        response, content = self.http.request(uri, method='GET', headers=headers, auth_without_challenge=True)
        logging.debug('response: %s', response)
        logging.debug('content: %s', content)

        tree = komcenter.get_tree_or_raise_error(response, content)

        entries = []
        for record in tree.findall('isg'):
            entries.append({
                'id': record.find('id').text,
                'name': record.find('name').text,
                'email': record.find('email').text,
            })
        return entries



    #---------- Help ----------

    def help_examples(self):
        """usage examples"""
        doc = """Examples:
    ${name} CONTEXT COMMAND ARGS

  View IP Entries:
    # get entry by ip
    ${name} ip get fufu.ethz.ch
    # get all entries associated with ip
    ${name} ip get 129.132.219.236
    # get entry for given ip/hostname pairs
    ${name} ip get 129.132.219.236/fufu.ethz.ch
    ${name} ip get 129.132.219.236/fofo.ethz.ch
    # get entry by mac
    ${name} ip get 00:1B:FC:A2:C3:9F

  Add an IP Entry:
    ${name} ip add 129.132.12.0 ipctl.icos.ethz.ch reverse:Y ttl:4800
    ${name} ip add 129.132.12.0 ipctl.icos.ethz.ch r:Y t:4800
    ${name} ip add 129.132.12.0 ipctl.icos.ethz.ch r:Y ip:129.132.12.49

  Change an IP Entry:
    # change entry for hostname
    ${name} ip set fufu.ethz.ch ttl:600
    # explicitly select hostname
    ${name} ip set 129.132.219.236/fofo.ethz.ch t:4800
    # change entry for ip, will prompt for hostname if there is more then one hostname bound to this ip
    ${name} ip set 129.132.219.236 t:4800
    # change ip of hostname
    ${name} ip set 129.132.219.236/fofo.ethz.ch ip:129.132.219.237

  Delete an IP Entry:
    # delete all entries associated with ip
    ${name} ip delete 129.132.12.49
    # delete entry for given ip/hostname pair
    ${name} ip delete 129.132.12.49/ipctl.icos.ethz.ch
    # delete entry by hostname
    ${name} ip delete ipctl.icos.ethz.ch

  View other IP Entries:
    ${name} ip free 129.132.12.0
    ${name} ip used 129.132.12.0


  View an Alias Entry:
    ${name} alias get alias.icos.ethz.ch

  Add an Alias Entry:
    ${name} alias add ipctl.icos.ethz.ch alias.icos.ethz.ch ttl:4800

  Change an Alias Entry:
    ${name} alias set alias.icos.ethz.ch ttl:600

  Delete an Alias Entry:
    ${name} alias delete alias.icos.ethz.ch

  View other Alias Entries:
    ${name} alias all ipctl.icos.ethz.ch
        """
        #doc = self._help_reindent(doc)
        doc = self._help_preprocess(doc, None)
        doc = doc.rstrip() + '\n' # trim down trailing space
        self.stdout.write(self._str(doc))
        self.stdout.flush()

    #---------- Overriden methods ----------

    _contexts = {
        'ip': ('add', 'delete', 'get', 'set', 'free', 'used'),
        'alias': ('add', 'delete', 'get', 'set', 'all'),
    }
    def precmd(self, argv):
        """Complete abbreviated command names."""
        if argv:
            args = argv[:]
            #assert len(args) >=2, 'Wrong number of arguments. Expected at least 2, got %d' % len(argv)
            try:
                context = util.get_canonical_name(args.pop(0), ('ip', 'alias'), label='context')
                logging.debug('canonical context name: %s', context)
                action = util.get_canonical_name(args.pop(0), self._contexts[context], label='action')
                logging.debug('canonical action name: %s', action)
                cmdname = self._get_canonical_cmd_name('%s_%s' % (context, action))
                logging.debug('canonical command name: %s', cmdname)
                if cmdname:
                    argv = [cmdname]
                    argv.extend(args)
            except IndexError, e:
                raise cli.CmdlnUserError('Wrong number of arguments. Expected at least 2, got %d' % len(argv))
            except util.UnknownNameError, e:
                pass
        return argv

    def _get_cmds_data(self):
        """Split command name into context and action for help messages."""
        cmds_data = super(IpCtl, self)._get_cmds_data()
        for index,data in enumerate(cmds_data):
            cmd_name = data[0]
            cmd_help = data[1]
            cmd_label = self._get_cmd_label(cmd_name)
            if cmd_name != cmd_label:
                cmds_data[index] = (cmd_label, cmd_help)
        return cmds_data

    def _get_cmd_label(self, cmd_name):
        if cmd_name:
            maybe_context = cmd_name.split('_', 1)[0]
            if maybe_context in self._contexts.keys():
                cmd_name = cmd_name.replace('_', ' ', 1)
        return cmd_name

    def _help_preprocess_cmd_name(self, help, cmdname=None):
        """FIXME: all this is way to hacky :-("""
        marker = "${cmd_name}"
        handler = self._get_cmd_handler(cmdname)
        if not handler:
            raise CmdlnError("cannot preprocess '%s' into help string: "
                             "could not find command handler for %r"
                             % (marker, cmdname))
        cmd_label = self._get_cmd_label(cmdname)
        s = cmd_label
        if hasattr(handler, "aliases"):
            s += " (%s)" % (", ".join(handler.aliases))
        help = help.replace(marker, s)
        return help


def main():
    try:
        sys.exit(IpCtl().main())
    except komcenter.KomcenterAuthenticationError, e:
        print 'Error:', e
        sys.exit(10)
    except Error, e:
        print 'Error:', e
        sys.exit(1)
    except KeyboardInterrupt:
        # FIXME: if in interactive shell abort current operation instead of exiting
        sys.exit(0)


if __name__ == '__main__':
    main()

