#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Query the ETHZ LDAP and Active Directory servers for information about users.
#
# (c) 2009, Steven Armstrong, ETH Zurich
#

__version__ = '0.1.3'

import sys
import re

# Configure logging
import logging
logging.basicConfig(level=logging.INFO, format='%(levelname)s: %(message)s')

from ethz import cli
from ethz.util import ldaplib

#LDAP_OPTIONS = { 'OPT_X_TLS_CACERTFILE': op.join(BASE_DIR, 'etc/ssl/certs/id.crt') }
#ldaplib.set_ldap_options(LDAP_OPTIONS)
ldap_url = 'ldaps://ldaps01.ethz.ch:636 ldaps://ldaps02.ethz.ch:636 ldaps://ldaps03.ethz.ch:636'


def encode_strings(entry, codec='latin1'):
    for k,v in entry.items():
        if not isinstance(v, unicode):
            s = unicode(v, 'utf-8')
            entry[k] = s.encode(codec)
    return entry



room_re = re.compile('(([^\d\s]+)\s?([^\d\s]{1})\s?(\d+(?:.\d+)*))\$')
def get_room_from_address(address):
    """
    postalAddress: Comput. Systems Biology, Stelling$ETH Zuerich, CAB J 71.5$Universitaetstrasse 6$8092 Zuerich
    """
    # FIXME: doesn't work with fancy addresses
    room = {'building': None, 'floor': None, 'number': None, 'fullname': 'unknown'}
    if not address:
        return room

    if 'Basel' in address:
        room['fullname'] = 'Basel'
        return room

    m = room_re.search(address)
    if m:
        groups = m.groups()
        room['building'] = groups[1]
        room['floor'] = groups[2]
        room['number'] = groups[3]
        room['fullname'] = '%s %s %s' % (room['building'], room['floor'], room['number'])
    return room



def get_user_details(uid, basedn='ou=users,ou=nethz,ou=id,ou=auth,o=ethz,c=ch', attributes=None):
    url = ldap_url
    binddn = 'cn=inf_proxy,ou=admins,ou=inf,ou=auth,o=ethz,c=ch'
    bindpw = 'inf_proxy'
    scope = ldaplib.ldap.SCOPE_SUBTREE
    filter = '(cn=%s)' % uid
    for entry in ldaplib.ldap_search(url, binddn, bindpw, basedn, scope=scope, filter=filter, attributes=attributes, async=True):
        entry = ldaplib.dict_from_ldap_result(entry[0])
        if 'postalAddress' in entry:
            entry['room'] = get_room_from_address(entry['postalAddress'])['fullname']
        else:
            entry['room'] = ''
        yield entry



def get_users_in_dn(basedn='ou=users,ou=nethz,ou=id,ou=auth,o=ethz,c=ch', attributes=None):
    url = ldap_url
    binddn = 'cn=inf_proxy,ou=admins,ou=inf,ou=auth,o=ethz,c=ch'
    bindpw = 'inf_proxy'
    scope = ldaplib.ldap.SCOPE_ONELEVEL
    filter = '(objectclass=posixAccount)'
    attrs = ['uid']
    for entry in ldaplib.ldap_search(url, binddn, bindpw, basedn, scope=scope, filter=filter, attributes=attrs, async=True):
        uid = entry[0][1]['uid'][0]
        for u in get_user_details(uid, 'ou=users,ou=nethz,ou=id,ou=auth,o=ethz,c=ch', attributes):
            yield u



def search_by_uid(uid):
    if not isinstance(uid, (list, tuple)):
        uid = [uid]
    basedn = 'ou=users,ou=nethz,ou=id,ou=auth,o=ethz,c=ch'
    attributes = ['cn', 'givenName', 'sn', 'gecos', 'telephoneNumber', 'mail', 'postalAddress', 'PERSID', 'uidNumber']
    for u in uid:
        for entry in get_user_details(u, basedn, attributes):
            yield entry



def search_by_ou(ou):
    if not isinstance(ou, (list, tuple)):
        ou = [ou]
    basedn = 'ou=inf,ou=auth,o=ethz,c=ch'
    attributes = ['cn', 'givenName', 'sn', 'gecos', 'telephoneNumber', 'mail', 'postalAddress', 'PERSID', 'uidNumber']
    for o in ou:
        # @group
        # @group@class
        parts = o.split('@')[1:]
        parts.reverse()
        dn = [basedn]
        while parts:
            dn.append('ou=%s' % parts.pop())
        dn.append('ou=users')
        dn.reverse()
        dn = ','.join(dn)
        for entry in get_users_in_dn(dn, attributes):
            yield entry



def search_by_name(names):
    url = ldap_url
    binddn = 'cn=inf_proxy,ou=admins,ou=inf,ou=auth,o=ethz,c=ch'
    bindpw = 'inf_proxy'
    scope = ldaplib.ldap.SCOPE_ONELEVEL
    basedn = 'ou=users,ou=nethz,ou=id,ou=auth,o=ethz,c=ch'
    attributes = ['cn', 'givenName', 'sn', 'gecos', 'telephoneNumber', 'mail', 'postalAddress', 'PERSID', 'uidNumber']
    attrs = ['uid']
    filter = ['(&'];
    for name in names:
        # approx match, slow
        #q = '(|(sn~=%(arg)s)(givenName~=%(arg)s)(mail~=%(arg)s))'
        q = '(|(sn=*%(name)s*)(givenName=*%(name)s*))'
        filter.append(q % {'name':name})
    filter.append(')')
    filter = ''.join(filter)
    for entry in ldaplib.ldap_search(url, binddn, bindpw, basedn, scope=scope, filter=filter, attributes=attrs, async=True):
        uid = entry[0][1]['uid'][0]
        for u in get_user_details(uid, basedn, attributes):
            yield u



class UserInfo(cli.Command):
    """Usage:
        ${name} [ARGS...]
        ${name} COMMAND [ARGS...]
        ${name} help COMMAND

    Query the ETHZ LDAP servers for information about users.

    ${option_list}
    ${command_list}
    ${help_list}
    """
    name = 'userinfo'
    option_interactive = True


    #---------- Commands ----------

    def do_uid(self, subcmd, opts, *args):
        """search by uid"""
        self.log.info('Searching by uid: %s', args)
        results = search_by_uid(args)
        self.process_results(results)


    def do_ou(self, subcmd, opts, *args):
        """search by organizational unit"""
        self.log.info('Searching by OU: %s', args)
        results = search_by_ou(args)
        self.process_results(results)


    def do_name(self, subcmd, opts, *args):
        """search by user name"""
        self.log.info('Searching by name: %s', args)
        results = search_by_name(args)
        self.process_results(results)


    @cli.alias('q')
    @cli.alias('exit')
    def _do_quit(self, argv):
        """hidden quit command to leave interactive shell"""
        sys.exit(0)


    #---------- Help ----------

    def help_examples(self):
        """usage examples"""
        doc = """Examples:
    ${name} COMMAND ARGS

    ${name} uid userid
    ${name} ou @groupname
    ${name} name First Last

If COMMAND is ommited it is guessed by inspecting the arguments:
    ${name} userid -> ${name} uid userid
    ${name} @groupname -> ${name} ou @groupname
      """
        doc = self._help_reindent(doc)
        doc = self._help_preprocess(doc, None)
        doc = doc.rstrip() + '\n' # trim down trailing space
        self.stdout.write(self._str(doc))
        self.stdout.flush()


    #---------- Overriden methods ----------

    _autodetect_patterns = {
        # asteven
        'uid': re.compile(r'^[a-z0-9-_.]+$', re.IGNORECASE),
        # @gonnet, @gonnet@stud
        'ou': re.compile(r'^(?:@[a-z0-9-_.]+){1,2}$', re.IGNORECASE),
    }
    def precmd(self, argv):
        """If no explicit command name was given try guessing it by inspecting the arguments"""
        if argv:
            cmdname = self._get_canonical_cmd_name(argv[0])
            if not cmdname:
                arg = argv[0]
                for name, pattern in self._autodetect_patterns.items():
                    if pattern.search(arg):
                        argv.insert(0, name)
                        break
        return argv


    #---------- Utilities ----------

    def process_results(self, results):
        for r in results:
            entry = dict.fromkeys(['cn', 'givenName', 'sn', 'gecos', 'telephoneNumber', 'mail', 'postalAddress', 'PERSID'], '')
            entry.update(r)
            entry = encode_strings(entry)

            try:
                format = ['%(cn)s', '%(uidNumber)s', '%(givenName)s', '%(sn)s', '%(telephoneNumber)s', '%(room)s', '%(mail)s']
                msg = '\t'.join(format) % entry
                print msg
            except KeyError, e:
                print 'Error:', e, entry



def main():
    try:
        sys.exit(UserInfo().main())
        # FIXME: user friendly error handling
    except KeyboardInterrupt:
        # FIXME: if in interactive shell abort current operation instead of exiting
        sys.exit(0)


if __name__ == '__main__':
    main()


