import sys
import os
import string
import re
import logging


def uniq(l):
    """make a list uniq while preserving order"""
    s = {}
    return [s.setdefault(e,e) for e in l if e not in s]



def _expand_range(s):
    """_expand_range(range_pattern) -> list of strings"""
    included = []
    excluded = []
    parts = s.split(',')
    for segment in parts:
        if segment.startswith('!'):
            segment = segment[1:]
            collection = excluded
        else:
            collection = included
        delim = None
        if '-' in segment:
            delim = '-'
        elif '..' in segment:
            # bash style
            delim = '..'
        if delim:
            start,end = segment.split(delim, 1)
            if start.startswith('0'):
                # pad with leading zeros
                format = '%%0%dd' % len(start)
            else:
                format = '%d'
            collection.extend([format % i for i in xrange(int(start), int(end)+1)])
        else:
            collection.append(segment)
    range = [x for x in included if x not in excluded]
    return range

def _merge_range(prefix, range, suffix):
    """_merge_range(prefix, range_pattern, suffix) -> list of strings"""
    return ['%s%s%s' % (prefix, r, suffix) for r in _expand_range(range)]

range_re = re.compile(r'([^{]*){([^{]*)}([^{]*)', re.IGNORECASE)
def expand_range_pattern(s):
    """expand_range_pattern(pattern) -> list of strings
    
    >>> expand_range_pattern('a{1-3}b')
    ['a1b', 'a2b', 'a3b']
    
    >>> expand_range_pattern('a{5..6}b')
    ['a5b', 'a6b']
    
    >>> expand_range_pattern('a{1,2,3-5}b')
    ['a1b', 'a2b', 'a3b', 'a4b', 'a5b']
    
    pad with leading zeros
    >>> expand_range_pattern('a{01-3}b')
    ['a01b', 'a02b', 'a03b']
    
    >>> expand_range_pattern('a{1-3}b{4,5}')
    ['a1b4', 'a1b5', 'a2b4', 'a2b5', 'a3b4', 'a3b5']

    >>> expand_range_pattern('a{1-5,!2-3}b{4,5}')
    ['a1b4', 'a1b5', 'a4b4', 'a4b5', 'a5b4', 'a5b5']

    >>> expand_range_pattern('a{1-3}b{4,5,!4}')
    ['a1b5', 'a2b5', 'a3b5']

    """
    matches = range_re.findall(s)
    if not matches:
        # assume just a single hostname
        return [s]
    else:
        names = []
        # popping from end of list is more efficient
        matches.reverse()
        while matches:
            parts = _merge_range(*matches.pop())
            if not names:
                names = parts
                continue
            tmp = []
            for name in names:
                for part in parts:
                    tmp.append('%s%s' % (name,part))
            names = tmp
        return names

#s = 'a{1-3}b{5..6}c{8-10,12}d'
#for n in expand_range_pattern(s):
#    print n



def get_ip_from_hostname(hostname):
    import socket
    import sys
    ipaddress = None
    try:
        try:
            ipaddress = socket.gethostbyname(hostname)
        except socket.gaierror, (code, message):
            print >>sys.stderr, 'get_ip_from_hostname:', code, message
    finally:
        return ipaddress



def get_username_password(username=None, host=None):
    logging.debug('get_username_password: username={0}, host={1}'.format(username, host))

    # check for username/password environment variables
    password = os.getenv('SANS_PASSWORD', default=None)
    # for username, a command line provided value has precedence
    if not username:
        username = os.getenv('SANS_USERNAME', default=None)

    if not username:
        # Take username from command line, prompt for password
        if host:
            try:
                import netrc
                n = netrc.netrc()
                authenticators = n.authenticators(host)
                if authenticators:
                    # Use credentials found in ~/.netrc file
                    username,password = authenticators[0],authenticators[2]
            except IOError:
                # no ~/.netrc file
                pass
        if not username:
            # Use username of the current user, prompt for password
            import getpass
            username = getpass.getuser()
    if not password:
        import getpass
        prompt = 'Password for %s' % username
        if host:
            prompt = '%s@%s' % (prompt, host)
        try:
            password = getpass.getpass(prompt='%s: ' % prompt, stream=sys.stderr)
        except TypeError: # unix_getpass() got an unexpected keyword argument 'stream'
            # Fallback for old python version
            password = getpass.getpass(prompt='%s: ' % prompt)

    return (username, password)



not_hexdigits = re.compile(r'[^%s]' % string.hexdigits)
def format_mac(mac, delimiter=':'):
    if not mac:
        return mac

    # nuke any existing delimiters
    mac = not_hexdigits.sub('', mac)

    # /me likes uppercase mac addresses
    mac = mac.upper()

    # re-format with given delimiter
    return delimiter.join([mac[i:i+2] for i in xrange(0, len(mac), 2)])


class SortedDict(dict):
    """
    A dictionary that keeps its keys in the order in which they're inserted.
    """
    def __new__(cls, *args, **kwargs):
        instance = super(SortedDict, cls).__new__(cls, *args, **kwargs)
        instance.keyOrder = []
        return instance

    def __init__(self, data=None):
        if data is None:
            data = {}
        elif isinstance(data, GeneratorType):
            # Unfortunately we need to be able to read a generator twice.  Once
            # to get the data into self with our super().__init__ call and a
            # second time to setup keyOrder correctly
            data = list(data)
        super(SortedDict, self).__init__(data)
        if isinstance(data, dict):
            self.keyOrder = data.keys()
        else:
            self.keyOrder = []
            for key, value in data:
                if key not in self.keyOrder:
                    self.keyOrder.append(key)

    def __deepcopy__(self, memo):
        return self.__class__([(key, deepcopy(value, memo))
                               for key, value in self.iteritems()])

    def __setitem__(self, key, value):
        if key not in self:
            self.keyOrder.append(key)
        super(SortedDict, self).__setitem__(key, value)

    def __delitem__(self, key):
        super(SortedDict, self).__delitem__(key)
        self.keyOrder.remove(key)

    def __iter__(self):
        return iter(self.keyOrder)

    def pop(self, k, *args):
        result = super(SortedDict, self).pop(k, *args)
        try:
            self.keyOrder.remove(k)
        except ValueError:
            # Key wasn't in the dictionary in the first place. No problem.
            pass
        return result

    def popitem(self):
        result = super(SortedDict, self).popitem()
        self.keyOrder.remove(result[0])
        return result

    def items(self):
        return zip(self.keyOrder, self.values())

    def iteritems(self):
        for key in self.keyOrder:
            yield key, self[key]

    def keys(self):
        return self.keyOrder[:]

    def iterkeys(self):
        return iter(self.keyOrder)

    def values(self):
        return map(self.__getitem__, self.keyOrder)

    def itervalues(self):
        for key in self.keyOrder:
            yield self[key]

    def update(self, dict_):
        for k, v in dict_.iteritems():
            self[k] = v

    def setdefault(self, key, default):
        if key not in self:
            self.keyOrder.append(key)
        return super(SortedDict, self).setdefault(key, default)

    def value_for_index(self, index):
        """Returns the value of the item at the given zero-based index."""
        return self[self.keyOrder[index]]

    def insert(self, index, key, value):
        """Inserts the key, value pair before the item with the given index."""
        if key in self.keyOrder:
            n = self.keyOrder.index(key)
            del self.keyOrder[n]
            if n < index:
                index -= 1
        self.keyOrder.insert(index, key)
        super(SortedDict, self).__setitem__(key, value)

    def copy(self):
        """Returns a copy of this object."""
        # This way of initializing the copy means it works for subclasses, too.
        obj = self.__class__(self)
        obj.keyOrder = self.keyOrder[:]
        return obj

    def __repr__(self):
        """
        Replaces the normal dict.__repr__ with a version that returns the keys
        in their sorted order.
        """
        return '{%s}' % ', '.join(['%r: %r' % (k, v) for k, v in self.items()])

    def clear(self):
        super(SortedDict, self).clear()
        self.keyOrder = []


